"""
Name: ACSNI-save
Author: Chinedu A. Anene, Phd
"""

from pandas import read_csv
import pandas as pd
import sys, os, random
from ACSNI.dbs import AcsniResults, save_acsni


def shuffle_exp(exp):
    """
    Shuffle expression matrix

    Parameters
    ----------
    exp: Original expression matrix

    Returns
    -------
    lm_w: Shuffled expression matrix
    """

    nam_cols = exp.select_dtypes(include=["number", "float", "int"]).columns
    if len(nam_cols) == 0:
        sys.exit("ERROR: Expression matrix must have numeric columns")

    new_exp = exp.select_dtypes(exclude=["number", "float", "int"]).copy()
    k_sam = len(new_exp.index)
    for i in nam_cols:
        new_exp[i] = random.sample(list(exp[i].values), k_sam)
    return new_exp

def save_merged_w_n_p_d(co, w, n, p, d, nn, run_info):
    """
    Save multi-run
    Parameters
    ----------
    w: weights
    n: networks
    p: predictions
    d: direction
    co: Pathway code
    nn: Name of input expression matrix
    run_info: Run parameters

    Returns
    -------
    """
    p_sum = p.select_dtypes(include=["number", "float", "int"]).copy()
    nnm = len(p_sum.columns)

    p["Sum_stat"] =  p_sum.astype(bool).sum(axis=1)
    p["Boot_Count"] = nnm

    dbs_results = AcsniResults(co=co, w=w, n=n, p=p, d=d,
                                   run_info=run_info)

    save_acsni(dbs_results, ("dbs" + nn[:-4] + ".ptl"))
    n1 = dbs_results.add_sub_process_direction(c=2)
    n1.to_csv("Network_{}".format(nn), index=False)
    return

def merge_multi(path, info, file):
    """
    Merge bootstrap run for summary start.
    Parameters
    ----------
    path: Path to the working directory
    info: Run information
    file: Input expression name

    Returns
    -------
    out: List of dataframes W, N, P, Co
    """
    files = [f for f in os.listdir(path) if f.endswith(".csv")]
    ww, nn, pp, dd, co = list(), list(), list(), list(), list()

    for i in files:
        if i.startswith("code"):
            co.append(read_csv(os.path.join(path, i)))
            os.remove(os.path.join(path, i))
        elif i.startswith("W"):
            ww.append(read_csv(os.path.join(path, i)))
            os.remove(os.path.join(path, i))
        elif i.startswith("N"):
            nn.append(read_csv(os.path.join(path, i)))
            os.remove(os.path.join(path, i))
        elif i.startswith("P"):
            pp.append(read_csv(os.path.join(path, i)))
            os.remove(os.path.join(path, i))
        elif i.startswith("D"):
            dd.append(read_csv(os.path.join(path, i)))
            os.remove(os.path.join(path, i))
        else :
            pass

    if len(co) == 1:
        co = co[0]

        ww = ww[0]
        ww.set_index('ID')

        nn = nn[0]
        nn.set_index('name')

        pp = pp[0]
        pp.set_index('name')

        dd = dd[0]
        dd.set_index('name')

    elif len(co) > 1:
        co = pd.concat(co, axis=0)

        ww = [i.set_index('ID') for i in ww]
        ww = pd.concat(ww, axis=1)

        nn = [i.set_index('name') for i in nn]
        nn = pd.concat(nn, axis=1)

        pp = [i.set_index('name') for i in pp]
        pp = pd.concat(pp, axis=1)

        dd = [i.set_index('name') for i in dd]
        dd = pd.concat(dd, axis=1)

    else:
        sys.exit("Fatal Error")

    save_merged_w_n_p_d(co=co, w=ww, n=nn, p=pp, d=dd, nn=file,
                        run_info=info)

    mf = [f for f in os.listdir(path) if f.endswith(".h5")]
    for m in mf:
        os.remove(os.path.join(path, m))

    return







"""
Name: ACSNI-core
Author: Chinedu A. Anene, Phd
"""

import sys
import pandas as pd
from ACSNI.mud import DimRed
from ACSNI.dat import get_row_index, filter_uninformative, \
    scale_first, get_model_weights
from ACSNI.network import NetScore


def check_input(nc):
    """
    Check number of samples

    Parameters
    ----------
    nc: Number of columns

    Returns
    -------
    end: Pass or early stop
    """

    if nc <= 15:
        sys.exit("{} samples is small for the tool"
                 .format(nc))
    elif nc >= 800:
        sys.exit("{} samples is large, use "
                 "ACSNI-split".format(nc))
    else:
        print("{} samples".format(nc))
    return

def check_name(e_df):
    """
    Check expression matrix

    Parameters
    ----------
    e_df: Expression matrix

    Returns
    -------
    end: Pass or early exit
    """
    c_col = len(e_df.columns)
    num_exp = e_df.select_dtypes(include=["number"]).copy()
    n_col = c_col - len(num_exp.columns)

    if n_col >= 2:
        print("ERROR: {} categorical columns found".format(n_col))
        sys.exit("Remove the {} extra label columns in "
                 "the expression matrix".format(n_col-1))
    else:
        pass
    return

def check_duplicates(df):
    """
    Check duplicated genes

    Parameters
    ----------
    df: Expression matrix

    Returns
    -------
    end: Pass or early exit
    """

    id_name = get_row_index(df)

    is_dup_row = df.duplicated(subset=id_name, keep='first')
    dup_genes = df[is_dup_row][id_name]

    if dup_genes.size >= 1:
        print("ERROR: Found {} duplicated genes".format(dup_genes.size))
        print(dup_genes)
        sys.exit("Remove duplicated genes and try again")
    else:
        pass
    return


def main_model_prediction(inp, d_list, gi, lp, s, run):
    """
    Run final model

    Parameters
    ----------
    inp: expression
    d_list: Prior
    gi: Gene set name
    s: Boot number
    lp: Layer dimension
    run: The details of the run to be performed

    Returns
    -------
    count_pre: predictions
    weights: functional units
    aede: optimal dimension
    """

    exp_mat = filter_uninformative(inp, mad_f=run.get("m"))
    x_train, y_train = scale_first(input_file=exp_mat, d_list=d_list)
    check_input(x_train.shape[1])

    if run.get("w") is None:
        w_nn = run.get("w")

    else:
        w_nn = get_model_weights(pd.read_csv(run.get("w")), y_train)

    mod = DimRed(x=x_train, w = w_nn, p = lp)
    mod.fit()
    mod.add_reduced_row(inp.select_dtypes(include=["number","float","int"]).columns)

    net = NetScore(x=exp_mat, w=mod.get_reduced(), p=run.get("c"))
    net.fit()
    net.save_net(str(s), gi, run.get("i"))

    if run.get("f") == 1:
        mod.get_reduced().to_csv("W{}_{}.csv".format(str(s), gi), index=False)

        net.get_predicted(run.get("f")).to_csv("P{}_{}_{}".format(str(s), gi,
                                                                  run.get("i")), index=False)
        count_p = mod.get_aede()

    elif run.get("f") == 0:
        count_p = net.get_predicted(run.get("f"))
        count_p.columns = ["name", gi]

    else:
        sys.exit("ERROR: Invalid value set for parameter -f")

    return count_p

def merge_minimal_out(x, nn):
    """
    Merge the minimal output results

    Parameters
    ----------
    x: Dictionary of results
    nn: Name of input file

    Returns
    -------
    end: Completion
    """

    mse = x[0]
    for i in x[1:]:
        mse = pd.merge(mse, i, how="outer", on="name")
    mse.to_csv("predicted_{}.csv".format(nn), index=False)
    return



"""
Name: ACSNI-database
Author: Chinedu A. Anene, Phd
"""

import pandas as pd
from sklearn.linear_model import LinearRegression,LogisticRegression
import numpy as np
import pickle
import sys


class AcsniResults:
  """
  Class for ACSNI results
  """

  def __init__(self, co, w, n, p, d, run_info):
      self.co = co
      self.w = w
      self.n = n
      self.p = p
      self.d = d
      self.__lm_w = None
      self.__nm = None
      self.run_info = run_info

  def __str__(self):
      n_boot = self.p["Boot_Count"][1]
      n_modules = int((len(self.w.columns)-1)/n_boot)
      return "ACSNI result with {} modules over {} bootstraps.".format(n_modules, n_boot)

  def __repr__(self):
      return "\n" + self.__str__()

  def get_co(self):
      return self.co

  def get_w(self):
      return self.w

  def get_n(self):
      return self.n

  def get_p(self):
      return self.p

  def get_d(self):
      return self.d

  def get_run_info(self):
      return self.run_info

  def get_global_prediction(self):
      gr = self.p.copy()
      gr["name"] = gr.index
      gr = pd.concat([self.p["name"],
                      self.p.select_dtypes(include=["number", "float", "int"]).copy()],
                     axis=1)
      return gr

  def get_n_boot_units(self):
      boot = self.p["Boot_Count"][1]
      mod = int((len(self.w.columns)-1)/boot)
      return boot, mod

  def get_associated_module(self, target):
      names_tar = target.loc[:, target.columns != "ID"].columns[0]
      tar = self.get_w()
      tar = pd.merge(tar, target, left_on="ID", right_on="ID", how="left")
      return tar, names_tar

  def get_sub_process_network(self, c):
      n_p = self.p.loc[self.p["Sum_stat"] >= c]
      n_p = n_p.select_dtypes(exclude=["number", "float", "int"]).copy()
      n_p = n_p[[col for col in n_p if col.startswith("AE")]]
      n_p["name"] = n_p.index
      n_p = n_p.melt(id_vars="name", var_name="sub", value_name="Predicted")
      n_p = n_p.loc[n_p["Predicted"] == "P"]
      n_p["Set"] = "w" + n_p["sub"].astype(str)
      n_p = n_p[["name", "sub", "Set"]]
      return n_p

  def add_sub_process_direction(self, c):
      n_p = self.get_sub_process_network(c)
      n_p["match"] = n_p["name"] + n_p["Set"]
      self.d["name"] = self.d.index
      d_p = self.d.melt(id_vars="name", var_name="sub", value_name="Direction")
      d_p["Set"] = "w" + d_p["sub"].astype(str)
      d_p["match"] = d_p["name"] + d_p["Set"]
      d_p = d_p.loc[d_p["match"].isin(n_p["match"])]
      d_p = d_p[["Direction", "match"]]
      n_p = pd.merge(n_p, d_p, how="left", on="match")
      return n_p[["name", "sub", "Direction"]]

  def get_explained_variation(self, t, c):
      """
      Get level of interaction between phenotype and subprocesses.

      """

      self.__nm = t.loc[:, t.columns != "ID"].columns[0]

      weights = self.w.select_dtypes(include=["number", "float", "int"]).columns
      weights = [cc for cc in weights if cc.startswith("AE")]
      lm_w = pd.DataFrame({"sub": weights})

      self.w = pd.merge(self.w, t, left_on="ID", right_on="ID", how="left")

      lm_g = []
      for i in weights:
          x = np.array(self.w[i]).reshape(-1, 1)
          y = self.w[[self.__nm]].values.ravel()

          if c == "character":
              lm_r = LogisticRegression(n_jobs=-1)
          elif c == "numeric":
              lm_r = LinearRegression(n_jobs=-1)
          else:
              sys.exit("Invalid value set for parameter -c. "
                       "Use numeric or character for the type of phenotype")

          lm_r.fit(x, y)
          lm_r.score(x, y)
          lm_g.append(lm_r.score(x, y))
      lm_g = pd.DataFrame(lm_g)
      self.__lm_w = pd.concat([lm_w, lm_g], axis=1)
      return

  def get_summary_stat_phenotype(self, t, c):
      """
      Phenotype association statistics
      Parameters
      ----------
      t: phenotype file
      c: Type of variable

      Returns
      -------
      results and name

      """

      self.get_explained_variation(t, c)
      qq = self.__lm_w[[0]].quantile([.25, .75])
      q25 = qq.loc[0.25, 0]
      q75 = qq.loc[0.75, 0]

      m = float(self.__lm_w[[0]].mean())
      sd = float(self.__lm_w[[0]].std())
      print("Statistics of variations in subprocesses explained by {}".format(self.__nm))
      print("q25 {} \n q75 {} \n mean {} \n std {}".format(q25, q75, m, sd))
      self.__lm_w["Association"] = np.where(self.__lm_w[0] >= q75, "Strong", "Weak")
      return self.__lm_w, self.__nm


class ACSNIDeriveResults:
  """
  Class for ACSNI Derive results
  """

  def __init__(self, ac, n, d, fd, co, run_info):
      self.n = n
      self.d = d
      self.ac = ac
      self.fd = fd
      self.co = co
      self.run_info = run_info

  def __str__(self):
      return "ACSNI Derive raw results"

  def __repr__(self):
      return "\n" + self.__str__()

  def get_run_info(self):
      return self.run_info

  def get_ac(self):
      return self.ac

  def get_n(self):
      return self.n

  def get_d(self):
      return self.d

  def get_fd(self):
      return self.fd

  def get_co(self):
      return self.co

def save_acsni(obj, filename):
    with open(filename, 'wb') as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)

#dbst = load_acsni("dbsAA_.csv.ptl")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as rq:
    install_requires = rq.read()


setuptools.setup(
    name="ACSNI",
    version="1.0.4",
    scripts=["bin/ACSNI-run", "bin/ACSNI-derive", "bin/ACSNI-get", "bin/ACSNI-split"],
    author="Chinedu A. Anene",
    collaborator="Faraz Khan",
    author_email="caanenedr@outlook.com",
    description="automatic context-specific network inference",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/caanene1/ACSNI",
    download_url = "https://github.com/caanene1/ACSNI/releases/download/1.0.4/ACSNI-1.0.4-py3-none-any.whl",
    packages=setuptools.find_packages(include=["ACSNI"]),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=install_requires,
    python_requires='>=3.6',
)

# Build >> python3 setup.py sdist bdist_wheel
# Upload >> sudo twine upload dist/*
